#include "camerafocus.h"
#include "ui_camerafocus.h"



CameraFocus::RefreshFrame::RefreshFrame(CameraFocus *_camera) : QThread() {
    camera = _camera;
}

void CameraFocus::RefreshFrame::run() {
    camera->refreshFrame();
}


CameraFocus::CameraFocus(SandboxSetup *sandbox, MonitorGui *_mg, QWidget *parent) :
    SubApp("Camera focus", "Error", parent),
    ui(new Ui::CameraFocus)
{
    ui->setupUi(this);
    setup = sandbox;
    mg = _mg;
    myThread = new RefreshFrame(this);
    blackScreen = new QtFullScreen(mg->getResolution(), true, this);
    setup->loadFrameProcessProfil();

    frameTimer = new QTimer(this);
    connect(frameTimer, &QTimer::timeout, this, &CameraFocus::startCapture);
}

CameraFocus::~CameraFocus()
{
    delete frameTimer;
    delete blackScreen;
    myThread->exit(0);
    myThread->wait();
    delete myThread;
    delete ui;
}

void CameraFocus::showEvent(QShowEvent *event){

    QWidget::showEvent(event);
    initCameraParams();
    cv::Mat black = cv::Mat(setup->getBeamer()->getHeight(), setup->getBeamer()->getWidth(), CV_8UC3, cv::Scalar(0, 0, 0));
    cv::line(black, cv::Point(black.cols/2, 0), cv::Point(black.cols/2, black.rows-1), cv::Scalar(255, 0, 0), 4);
    cv::line(black, cv::Point(0, black.rows/2), cv::Point(black.cols-1, black.rows/2), cv::Scalar(255, 0, 0), 4);
    blackScreen->editGeometry(mg->getResolution());
    blackScreen->imShow(black);
    frameTimer->start(100);
}

void CameraFocus::closeEvent(QCloseEvent *event){

    frameTimer->stop();
    blackScreen->close();
    QWidget::closeEvent(event);
}

// TODO : override "checkRoutine" to ensure the profil makes sense

void CameraFocus::valideRoutine(){

    endSuccess = true;
}

void CameraFocus::cancelRoutine(){

    endSuccess = false;
}

void CameraFocus::startCapture(){
    myThread->start();
}

void CameraFocus::refreshFrame(){

    setup->getCamera()->capture();
    FrameProcessProfil *profil = setup->getBeamer()->getProfil();

    cv::Mat gray, cannyGray, rgbFromGray, cannyRgb;
    cv::Mat rgb = setup->getCamera()->getColorFrame();
    std::vector<cv::Point3i> crc;

    if(profil->getCannyEdgeThreshold() > 0 && profil->getHoughAccThreshold() > 0){

        double minDist = rgb.cols*(double)profil->getMinDistance()/100;
        double minRadius = (profil->getMinRadius()>0) ? rgb.cols*(double)profil->getMinRadius()/100 : 0;
        double maxRadius = (profil->getMaxRadius()>0) ? rgb.cols*(double)profil->getMaxRadius()/100 : 0;
        crc = setup->getBeamer()->findCircles( rgb,
                                               profil->getContrast(),
                                               profil->getBrightness(),
                                               minDist,
                                               (int)profil->getCannyEdgeThreshold(),
                                               (int)profil->getHoughAccThreshold(),
                                               minRadius,
                                               maxRadius );
    }

    // Preview image for the user
    cv::cvtColor(rgb, gray, CV_BGR2GRAY);
    gray = setup->getBeamer()->editContrast(gray, (double)profil->getContrast(), (double)profil->getBrightness());
    cv::Canny(gray, cannyGray, profil->getCannyEdgeThreshold(), profil->getCannyEdgeThreshold()/2);

    cv::cvtColor(gray, rgbFromGray, CV_GRAY2BGR);
    cv::cvtColor(cannyGray, cannyRgb, CV_GRAY2BGR);

    if(!crc.empty()){
        cv::circle(rgbFromGray, cv::Point(crc[0].x, crc[0].y), crc[0].z, cv::Scalar(0,255,0), 3);
        cv::circle(cannyRgb, cv::Point(crc[0].x, crc[0].y), crc[0].z, cv::Scalar(0,255,0), 3);
    }

    QImage img = QImage((uchar *)rgbFromGray.data, (int)rgbFromGray.cols, (int)rgbFromGray.rows, static_cast<int>(rgbFromGray.step.buf[0]), QImage::Format_RGB888);
    QPixmap px = QPixmap::fromImage(img);

    QImage cannyImg = QImage((uchar *)cannyRgb.data, (int)cannyRgb.cols, (int)cannyRgb.rows, static_cast<int>(cannyRgb.step.buf[0]), QImage::Format_RGB888);
    QPixmap cannyPx = QPixmap::fromImage(cannyImg);

    ui->lblFrameColor->setPixmap(px);
    ui->lblFrameCanny->setPixmap(cannyPx);

}


void CameraFocus::initCameraParams(){

    FrameProcessProfil *profil = setup->getBeamer()->getProfil();
    loadProfil(profil, &defaultProfil);
}

void CameraFocus::loadProfil(FrameProcessProfil *profilLoaded, FrameProcessProfil *profilSaved){

    // save profil
    profilSaved->setContrast(profilLoaded->getContrast());
    profilSaved->setBrightness(profilLoaded->getBrightness());
    profilSaved->setMinDistance(profilLoaded->getMinDistance());
    profilSaved->setCannyEdgeThreshold(profilLoaded->getCannyEdgeThreshold());
    profilSaved->setHoughAccThreshold(profilLoaded->getHoughAccThreshold());

    bool isUnrestricted = profilLoaded->getContrast() > 5.00 ||
                          profilLoaded->getBrightness() < -1275 ||
                          profilLoaded->getHoughAccThreshold() > 300;

    ui->ckbxAdvanced->setChecked(isUnrestricted);
    switchMode(!isUnrestricted);

    // set values
    ui->sbxContrast->setValue((double)profilLoaded->getContrast());
    ui->sbxBrightness->setValue((int)profilLoaded->getBrightness());
    ui->sbxMinDistance->setValue((int)profilLoaded->getMinDistance());
    ui->sbxCannyThreshold->setValue((int)profilLoaded->getCannyEdgeThreshold());
    ui->sbxAccThreshold->setValue((int)profilLoaded->getHoughAccThreshold());

    // check for the known target's size
    if(profilLoaded->getMinRadius() > 0 && profilLoaded->getMaxRadius() > 0){
        profilSaved->setMinRadius((int)profilLoaded->getMinRadius());
        profilSaved->setMaxRadius((int)profilLoaded->getMaxRadius());
        ui->sbxMinRadius->setValue((int)profilLoaded->getMinRadius());
        ui->sbxMaxRadius->setValue((int)profilLoaded->getMaxRadius());
        ui->ckbxTargetSize->setChecked(true);
        ui->ckbxTargetSize->clicked(true);
    }else{
        profilLoaded->setMinRadius(0);
        profilLoaded->setMaxRadius(0);
        profilSaved->setMinRadius(0);
        profilSaved->setMaxRadius(0);
    }

}

void CameraFocus::on_btnReset_clicked()
{
    loadProfil(&defaultProfil, setup->getBeamer()-> getProfil());
}

void CameraFocus::on_sldContrast_sliderMoved(int position)
{
    double val = position/100.0;
    setup->getBeamer()->getProfil()->setContrast(val);
    ui->sbxContrast->setValue(val);
}

void CameraFocus::on_sbxContrast_valueChanged(double arg1)
{
    setup->getBeamer()->getProfil()->setContrast(arg1);
    ui->sldContrast->setValue((int)(arg1*100));
}

void CameraFocus::on_sldBrightness_sliderMoved(int position)
{
    setup->getBeamer()->getProfil()->setBrightness(position);
    ui->sbxBrightness->setValue(position);
}

void CameraFocus::on_sbxBrightness_valueChanged(int arg1)
{
    setup->getBeamer()->getProfil()->setBrightness(arg1);
    ui->sldBrightness->setValue(arg1);
}

void CameraFocus::on_sldMinDistance_sliderMoved(int position)
{
    setup->getBeamer()->getProfil()->setMinDistance((uint)position);
    ui->sbxMinDistance->setValue(position);
}

void CameraFocus::on_sbxMinDistance_valueChanged(int arg1)
{
    setup->getBeamer()->getProfil()->setMinDistance((uint)arg1);
    ui->sldMinDistance->setValue(arg1);
}

void CameraFocus::on_sldCannyThreshold_sliderMoved(int position)
{
    setup->getBeamer()->getProfil()->setCannyEdgeThreshold((uint)position);
    ui->sbxCannyThreshold->setValue(position);
}

void CameraFocus::on_sbxCannyThreshold_valueChanged(int arg1)
{
    setup->getBeamer()->getProfil()->setCannyEdgeThreshold((uint)arg1);
    ui->sldCannyThreshold->setValue(arg1);
}

void CameraFocus::on_sldAccThreshold_sliderMoved(int position)
{
    setup->getBeamer()->getProfil()->setHoughAccThreshold((uint)position);
    ui->sbxAccThreshold->setValue(position);
}

void CameraFocus::on_sbxAccThreshold_valueChanged(int arg1)
{
    setup->getBeamer()->getProfil()->setHoughAccThreshold((uint)arg1);
    ui->sldAccThreshold->setValue(arg1);
}

void CameraFocus::on_sldMinRadius_sliderMoved(int position)
{
    setup->getBeamer()->getProfil()->setMinRadius((uint)position);
    ui->sbxMinRadius->setValue(position);
}

void CameraFocus::on_sbxMinRadius_valueChanged(int arg1)
{
    setup->getBeamer()->getProfil()->setMinRadius((uint)arg1);
    ui->sldMinRadius->setValue(arg1);
}

void CameraFocus::on_sldMaxRadius_sliderMoved(int position)
{
    setup->getBeamer()->getProfil()->setMaxRadius((uint)position);
    ui->sbxMaxRadius->setValue(position);
}

void CameraFocus::on_sbxMaxRadius_valueChanged(int arg1)
{
    setup->getBeamer()->getProfil()->setMaxRadius((uint)arg1);
    ui->sldMaxRadius->setValue(arg1);
}

void CameraFocus::on_ckbxTargetSize_clicked(bool checked)
{
    std::vector<QWidget *> lst = { ui->lblMinRadius, ui->sldMinRadius, ui->sbxMinRadius,
                                   ui->lblMaxRadius, ui->sldMaxRadius, ui->sbxMaxRadius };
    for(QWidget *w: lst){
        w->setEnabled(checked);
    }

    uint min = (checked) ? (uint)ui->sbxMinRadius->value() : 0;
    uint max = (checked) ? (uint)ui->sbxMaxRadius->value() : 0;
    setup->getBeamer()->getProfil()->setMinRadius(min);
    setup->getBeamer()->getProfil()->setMaxRadius(max);
}


void CameraFocus::switchMode(bool isRestricted){

    if(isRestricted){
        ui->sldContrast->setMaximum(500);
        ui->sbxContrast->setMaximum(5.00);

        ui->sldBrightness->setMinimum(-1275);
        ui->sbxBrightness->setMinimum(-1275);

        ui->sldAccThreshold->setMaximum(300);
        ui->sbxAccThreshold->setMaximum(300);

    }else{
        // we assume the matrixe for the accumulators has a ratio of 1 with the captured frame
        setup->getCamera()->capture();
        cv::Mat rgb = setup->getCamera()->getColorFrame();
        int maxHough = rgb.cols*rgb.rows;

        ui->sldContrast->setMaximum(25500);
        ui->sbxContrast->setMaximum(255.00);

        ui->sldBrightness->setMinimum(-65025);
        ui->sbxBrightness->setMinimum(-65025);

        ui->sldAccThreshold->setMaximum(maxHough);
        ui->sbxAccThreshold->setMaximum(maxHough);
    }
}

void CameraFocus::on_ckbxAdvanced_clicked(bool checked)
{
    switchMode(!checked);
}
