#include "croppingmask.h"
#include "ui_croppingmask.h"

CroppingMask::CroppingMask(SandboxSetup *sandbox, MonitorGui *_mg, QWidget *parent) :
    SubApp("Cropping mask", "Error", parent),
    ui(new Ui::CroppingMask)
{

    setup = sandbox;
    mg = _mg;
    ui->setupUi(this);
    ui->txtInstructions->setAlignment(Qt::AlignJustify);
    maskEdit = new MaskEdit(ui->frame->geometry());
    ui->vLayout->addWidget(maskEdit, 1);

    blueScreen = new QtFullScreen(mg->getResolution(), true, this);

    timer = new QTimer(this);
    connect(timer, &QTimer::timeout, this, &CroppingMask::refreshFrame);
}

CroppingMask::~CroppingMask()
{
    delete blueScreen;
    delete maskEdit;
    delete ui;
}

void CroppingMask::valideRoutine(){

    timer->stop();
    setup->getCamera()->capture();
    cv::Size s = setup->getCamera()->getDepthFrame().size();
    cv::Point center(s.width / 2, s.height / 2);
    std::vector<cv::Point> rectPoints = getRectPoints();
    cv::Mat depthFrame = setup->getCamera()->getDepthFrame();

    // TODO : center should be defined by the quadrilateral from rectPoints
    setup->setupAdjustMatrix(rectPoints, center);
    setup->setupCroppingMask(rectPoints);
    setup->getProjection()->setDistanceTopSandbox(depthFrame.at<float>(center));

    endSuccess = true;
}

void CroppingMask::cancelRoutine(){
    timer->stop();
    endSuccess = false;
}

void CroppingMask::showEvent(QShowEvent *event){
    QWidget::showEvent(event);

    // show blue screen
    if(!blueScreenInitilized){
        blueScreenInitilized = true;
        blueScreen->editGeometry(mg->getResolution());
    }
    cv::Mat blueMat = cv::Mat(1, 1, CV_8UC3, cv::Scalar(0, 0, 255));
    blueScreen->imShow(blueMat);
    init();
    timer->start(100);
}

void CroppingMask::closeEvent(QCloseEvent *event){
    QWidget::closeEvent(event);
    blueScreen->close();
}

void CroppingMask::init(){

    setup->getCamera()->capture();
    cameraColoredFrame = setup->getCamera()->getColorFrame();

    // load from config
    if(rectPoints.empty()){
        if(!setup->loadCroppingMask()){
            cv::Rect mask = setup->getCamera()->getCroppingMask();
            std::vector<cv::Point2i> pts = { cv::Point2i(mask.x, mask.y),
                                             cv::Point2i(mask.x, mask.y+mask.height),
                                             cv::Point2i(mask.x+mask.width, mask.y+mask.height),
                                             cv::Point2i(mask.x+mask.width, mask.y) };
            for(cv::Point2i pt : pts){
                loadedMask.push_back(setup->getProjection()->revertRotatePixel(pt));
            }
            rectPoints = loadedMask;
        }
    }

    // no config found
    if(rectPoints.empty() || !maskValideInFrame(&cameraColoredFrame)){
            float y = cameraColoredFrame.size().height;
            float x = cameraColoredFrame.size().width;
            rectPoints = std::vector<cv::Point2i>{ cv::Point2i(1.0/4*x, 1.0/4*y), cv::Point2i(1.0/4*x, 3.0/4*y), cv::Point2i(3.0/4*x, 3.0/4*y), cv::Point2i(3.0/4*x, 1.0/4*y) };
    }

    maskEdit->updateFrame(&cameraColoredFrame, &rectPoints);
}

void CroppingMask::refreshFrame(){
    setup->getCamera()->capture();
    cameraColoredFrame = setup->getCamera()->getColorFrame();
    maskEdit->updateFrame(&cameraColoredFrame);
}

bool CroppingMask::maskValideInFrame(cv::Mat *rgb){

    if(rectPoints.empty())
        return false;

    for(uint i=0; i<rectPoints.size(); i++){
        if( !(( 0<=rectPoints[i].x && rectPoints[i].x<rgb->size().width ) &&
            ( 0<=rectPoints[i].y && rectPoints[i].y<rgb->size().height ) ))
            {
            return false;
        }
    }

    return true;
}


void CroppingMask::on_btnClear_clicked()
{
    if(cameraColoredFrame.size().height != 0 && cameraColoredFrame.size().width != 0){
        float y = cameraColoredFrame.size().height;
        float x = cameraColoredFrame.size().width;
        rectPoints = std::vector<cv::Point2i>{ cv::Point2i(1.0/4*x, 1.0/4*y), cv::Point2i(1.0/4*x, 3.0/4*y), cv::Point2i(3.0/4*x, 3.0/4*y), cv::Point2i(3.0/4*x, 1.0/4*y) };
        maskEdit->updateFrame(&rectPoints);
    }
}

void CroppingMask::on_btnReset_clicked()
{
    if(!loadedMask.empty()){
        rectPoints = loadedMask;
        maskEdit->updateFrame(&rectPoints);
    }
}
